#!/usr/bin/perl -w

#    tex_to_lp.pl - Converts specially formatted TeX files into lp_solve
#                    models.
#    Copyright (C) 2007-2008  Christopher J. Hazard
#    ncsu at cjhazard dot edu swap the first two

#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 3 of the License, or
#    (at your option) any later version.

#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


# This program looks for the first occurance of \begin{align} %Model in the
# TeX input file.  It expects a line containing the objective, another line
# consisting of \text{s.t.}, and then the constraints.
# The top-level tokens are:
#  \max \min   :  used to begin the objective
#  \sum_{range}:  sums everything within range.  range can include a list
#                  of expressions separated by commas evaluated in order.
#                  The token \in may be used to denote a variable taking
#                  on each value within a set.  Sets may be defined in the
#                  parameters.pl file, or may be an integer range in the form
#                  of x..y, or a sequence of x;y;z.  The accepted constraints include \leq, \geq,
#                  \neq,>, <, =.  Summations may be nested.
#  \leq \geq = :  used in constraints
#  + -         :  used to separate terms in constraints
#  number      :  constant in constraint
#  x_{subscr}  :  denotes a variable with a specific subscript.  The
#                  subscript may be a list of single variables, and may also
#                  include addition and subtraction if included in
#                  parenthesis.  If the subscripted variable is defined in
#                  parameters.pl, then it may be used as a set in iteration.
#  \forall     :  used to denote a parameterization of the constraint.  The
#                  text following \forall is the same as the subscript of a
#                  \sum_{range}.
#  int free    :  used for defining integer variables as a constraint.
#                 E.g. x_{a} \in int \forall a \in A
#
# Also see example.tex and parameters.pl for example usage.
# Note that this program is not robust against non-unix newline standards.
#

use strict;

#grab command line parameters
my ($input_tex, $parameters_file) = @ARGV;
print "usage: tex_to_lp.pl input.tex [parameters.pl]" and exit unless $input_tex;
$parameters_file ||= 'parameters.pl';
my $output_lp = $input_tex;
$output_lp =~ s/\.tex$/\.lp/;

#open file
open FH, $input_tex or die "could not open file '$input_tex'";

my @lines = <FH>;
close FH;
chomp @lines;

#get constants/data about the problem
my $global_binds = do $parameters_file;

#pull model lines into @model
my @model;
my $in_model = 0;
for my $l (@lines) {
  if($in_model) {
    last if $l =~ /^\\end\{align/;
    push @model, $l;
    next;
  }
  $in_model = 1 if $l =~ /^\\begin\{align.*%Model$/;
}

#get rid of comments
$_ =~ s/\%.*//og foreach @model;
#get rid of any amphersands
$_ =~ s/\&//og foreach @model;
#get rid of labels
$_ =~ s/\\label\{.*?\}//og foreach @model;
#get rid of \notag
$_ =~ s/\\notag//og foreach @model;
#shrink spaces
$_ =~ s/\s+/ /og foreach @model;
#change \right(, etc. to just (
$_ =~ s/\\right//og foreach @model;
$_ =~ s/\\left//og foreach @model;

#get rid of \\'s, but if a line doesn't have \\, then concatenate on the next one
my $i = 0;
while($i < @model) {
  ++$i and next if $model[$i] =~ s/\\\\//og;
  #didn't have a double backslash, so concatenate the next line, and take it out of the model
  last unless exists $model[$i+1];
  $model[$i] .= splice @model, $i+1, 1;
}

#takes in an array of hash refs to binds, and returns a new
# bind combining both, leaving the parameters intact
sub CombineBinds {
  my %curbinds;
  @curbinds{keys %$_} = values %$_ foreach @_;
  return \%curbinds;
}

#pass in a reference to a string
#the function eats the variable name from that part of the string,
# and returns the translated result
sub TranslateLaTexVariableName {
  my $sref = shift;

  #grab name (or number)
  $$sref =~ s/^([a-zA-Z0-9]|[0-9]+)//;
  my $name = $1;
  #see if uses special tick mark to differentiate variable
  if($$sref =~ s/\^\\?([a-zA-Z0-9]+)//) {
    $name .= "__$1";
  }
  return $name;
}

#takes in a variable string and reference to variable binding
# returns an array, first element is the variable name,
# followed by the value of each subscript
sub TokenizeVariable {
  my ($name, $binds) = @_;

  my $alpha_name = TranslateLaTexVariableName \$name;

  #find any subscripts
  my @subscripts;
  if($name =~ s/_//) {
    #keep processing til out of subscript
    while($name) {
      next if $name =~ s/^(\{|\(|\))//; #skip open curly braces and parenthesis
      last if $name =~ /^\}/; #if final curly brace

      #skip whitespace
      next if $name =~ s/^\s+//;

      #if operation then store it
      if($name =~ s/^(\+|\-)//) {
	push @subscripts, $1;
	next;
      }

      my $tempname = TranslateLaTexVariableName \$name;

      #get subscript constant (it should be a constant)
      $tempname = GetVariableValue($tempname, $binds);

      #check to see if last subscript was operation
      if(@subscripts and
	 (($subscripts[-1] eq '+') or ($subscripts[-1] eq '-') )) {

	#perform the operation on the subscript stack
	my $op = delete $subscripts[-1];
	my $prev_val = delete $subscripts[-1];
	$tempname = eval "$tempname $op $prev_val";
      }

      #keep the new subscript value
      push @subscripts, $tempname;
    }
  }
  return ($alpha_name, @subscripts);
}

#pass in variable name string and a hash ref to the current
# bind scope, and it returns the value of the variable,
# or undef if it is not defined
sub GetVariableValue {
  my ($name, $binds, $subscripts) = @_;

  if(exists $binds->{$name}) {
    my $r = $binds->{$name};

    #traverse subscripts
    if($subscripts) {
      $r = $r->[int $_] foreach @$subscripts;
    }

    #if that variable isn't marked, then it's 0
    return 0 unless defined $r;
    return $r;
  }

  #check to see if it's a number
  if($name =~ m/^[0-9]/) {
    #see if integer range
    return [$1..$2] if $name =~ m/^([0-9]+)\.\.([0-9]+)$/;
    #see if semi-colon separated list
    return [split ';', $name] if $name =~ m/;/;

    #just return the number
    return int $name;
  }

  return undef;
}

#pass in variable string and binding list, and returns
# the specified variable or constant
sub BindVariable {
  my ($name, $binds) = @_;

  my ($varname, @subscripts) = TokenizeVariable($name, $binds);

  #if it's a constant, return the constant
  my $v = GetVariableValue($varname, $binds, \@subscripts);
  return $v if defined $v;

  #it's not a constant, so name the variable based on subscripts
  return join '_', ($varname, @subscripts);
}

#pass in a quantifier string and binding list
# returns an array of hashes that indicate variable bindings
sub QuantifiersToBindingList {
  my ($quantifier_string, $binds) = @_;
  return [{}] unless $quantifier_string;

  #contains an array of possible bindings
  my @bindlist = ({});

  #process every term in the quantifier list
  my @terms = split ',', $quantifier_string;
  foreach my $t (@terms) {
    $t =~ s/^\s+//;  #eat leading spaces
    my ($p1, $op, $p2) = split /\s+/, $t;

    if($op eq '\in') {
      #get set of $p2
      my $s = GetVariableValue($p2, $binds);
      die "undefined variable $p2" unless defined $s;

      my $translated_p1 = BindVariable($p1, $binds);

      @bindlist = map {
	my $b = $_;
	map {
	  my %h; #make a copy of the binds
	  @h{keys %$b} = values %$b;
	  $h{$translated_p1} = $_; #add the new bind
	  \%h; #use this new list of var binds
	} @$s; #add each new bind
      } @bindlist; #for every old bind

    } else { #constraint/comparison
      my %opmap = ( '\geq' => '>=',
		    '\leq' => '<=',
		    '\neq' => '!=',
		    '>' => '>',
		    '<' => '<',
		    '=' => '==', );

      #only keep those in the bind list that satisfy the comparison
      # for the given values
      @bindlist = grep {
	#need to combine the current summation bind with the outer scope binds passed in
	my $combined_binds = CombineBinds($_, $binds);
	#evaluate the condition
	eval(BindVariable($p1, $combined_binds)
	     . $opmap{$op} .
	     BindVariable($p2, $combined_binds)) } @bindlist;
    }

  }

  return \@bindlist;
}


my $cur_operation; #either + or -. how to join summations
my $cur_constant; #current multiplier (empty means 1)
my @summation_quantifiers;
my @variable_multiply_queue; #array of arrays of variables to multiply element-wise
my @dest; #new string to build

sub EmitQueue {
  #write out variables' dot products if have them
  if(@variable_multiply_queue > 0) {

    #for each variable-constant pair
    foreach my $i (0..$#{$variable_multiply_queue[0]}) {

      #gather all variables to be multiplied
      my @m = map {$_->[$i]} @variable_multiply_queue;

      #find out which one is the variable (the rest should be constants)
      my @vars = grep /^[a-zA-Z]/, @m;
      my @consts = grep !/^[a-zA-Z]/, @m;

      die "non-linear problem: too many variables: @vars" unless @vars <= 1;

      #combine constants
      my $const = (length $cur_constant > 0 ? $cur_constant : 1);
      $const *= $_ foreach @consts;

      #add on the new terms
      if($const) { #don't want to emit things that are 0
	push @dest, $cur_operation;
	push @dest, $const;
	#bind variable using all the available binds, as long as there's a variable
	push @dest, $vars[0] if @vars == 1;
      }
    }

    @variable_multiply_queue = ();
    $cur_constant = '';
    return;
  }

  #write out stored constant if have one
  push @dest, $cur_operation, $cur_constant if length $cur_constant > 0;
  $cur_constant = '';
}

#pass in a string, returns the string changed into an lp_solve formula
# optional second parameter is a hash, with specified variable bindings
sub GenerateEquation {
  #$src is the source of the equation,
  #$eqn_binds are the bindings for the equation
  my ($src, $eqn_binds) = @_;

  $cur_operation = '+';
  $cur_constant = '';
  @summation_quantifiers = ();
  @variable_multiply_queue = ();
  @dest = ();

  my %base_binds; #equation and global binds combined
  @base_binds{keys %$global_binds} = values %$global_binds;
  @base_binds{keys %$eqn_binds} = values %$eqn_binds;
  
  #set to the size of @dest when a condition is encountered to see if there is anything on the right side
  my $dest_size_at_condition = 0;

  while($src) {
    $src =~ s/^\s+// and next; #skip whitespace

    #take operation (additions and subtractions)
    if($src =~ s/^([\+\-])//) {
      EmitQueue;
      $cur_operation = $1;

      @summation_quantifiers = ();
      next;
    }

    #take comparison operations
    if($src =~ s/^(\\leq|\\geq|=)//) {
      EmitQueue;
      $cur_operation = '+';

      #if nothing on the one side of the constraint, put a 0
      push @dest, '+', '0', if @dest == 0;

      push @dest, '>=' if $1 eq '\geq';
      push @dest, '<=' if $1 eq '\leq';
      push @dest, '=' if $1 eq '=';

      $dest_size_at_condition = @dest;
      @summation_quantifiers = ();
      next;
    }

    #take summations
    if($src =~ s/^\\sum_\{(.*?)\}//) {
      push @summation_quantifiers, $1;
      next;
    }

    #take constants
    if($src =~ s/^([0-9]+)//) {
      $cur_constant = $1;
      next;
    }

    #take type definitions
    if($src =~ s/^\\in\s+(int|free)//) {
      #generate output and cut down to only the 'int' or 'free' statement and the variable
      my $qualifier = $1;
      EmitQueue;
      @dest = ($qualifier, $dest[-1]);
      last;
    }

    #take variables (with or without underscore, with or without {}'s)
    if($src =~ s/^(\\?[a-zA-Z][a-zA-Z0-9]*(\^\\?[a-zA-Z0-9]+)?(_\{.*?\}|_.*?)?)//) {
      my $var_string = $1;

      #if no additional bindings, write out variable once
      my @additional_binds;
      if(length @summation_quantifiers == 0) {
	@additional_binds = ({}) unless @additional_binds;
      }
      else {
	@additional_binds = 
	  @{QuantifiersToBindingList(
				     join(',', @summation_quantifiers),
				     \%base_binds)};
      }

      #make sure have new variables to add
      if(@additional_binds > 0) {
	my @new_vars = map {
	  BindVariable($var_string, CombineBinds(\%base_binds, $_))
	} @additional_binds;

	push @variable_multiply_queue, \@new_vars;
      }
      next;
    }

    #otherwise error
    $src =~ s/^([^\s]+)//;
    print "error, unrecognized token: $1 in $src\n";
    exit;
  }

  EmitQueue;
  
  #if nothing on the right side of the constraint, put a 0
  push @dest, '+', '0', if $dest_size_at_condition == @dest;
  
  my $retval = join ' ', @dest;
  #make sure the constraint has at least one variable, otherwise don't emit it
  return $retval if $retval =~ /[a-zA-Z]/;
  return '';
}

#pass in a string of a set of quantified constraints, and it will return an array of all constraints
sub GenerateConstraints {
  my ($expression, $quantifier) = split /\\forall/, shift;

  my $bindings = QuantifiersToBindingList($quantifier, $global_binds);
  my @constraints;
  push @constraints, GenerateEquation($expression, $_) foreach @$bindings;
  return \@constraints;
}

#actually do the conversion

open FH, ">$output_lp";

#convert objective function
$model[0] =~ s/\s*\\(max|min)\s+//;
print FH "$1: ";
print FH GenerateEquation($model[0], $global_binds), ";\n";

#assume that $model[1] is \text{s.t.}

#convert constraints
foreach my $m (@model[2..$#model]) {
  #print out all constraints that aren't blank
  $_ and print FH "$_;\n" foreach @{GenerateConstraints($m)};
}

close FH;

package DB;
sub printstack {
	my $i = 1;
	for (;;) {
		my ($package, $filename, $line, $subroutine) = caller($i);
		last if (!defined(scalar caller($i))); 
		#print "Subroutine: $subroutine; Line: $line; Arg: ${DB::args[0]}\n";
		print "Subroutine: $subroutine; Line: $line;\n";
		$i++;
	}
}
